/* ***** BEGIN LICENSE BLOCK *****
 * Version: MPL 1.1/GPL 2.0/LGPL 2.1
 *
 * The contents of this file are subject to the Mozilla Public License Version
 * 1.1 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is Annozilla.
 *
 * The Initial Developer of the Original Code is
 * Doug Daniels <rainking@alumni.rice.edu>
 * Portions created by the Initial Developer are Copyright (C) 2003
 * the Initial Developer. All Rights Reserved.
 *
 * Contributor(s):
 *  C. Greg Hagerty <cgreg@cgreg.com>
 *
 * Alternatively, the contents of this file may be used under the terms of
 * either the GNU General Public License Version 2 or later (the "GPL"), or
 * the GNU Lesser General Public License Version 2.1 or later (the "LGPL"),
 * in which case the provisions of the GPL or the LGPL are applicable instead
 * of those above. If you wish to allow use of your version of this file only
 * under the terms of either the GPL or the LGPL, and not to allow others to
 * use your version of this file under the terms of the MPL, indicate your
 * decision by deleting the provisions above and replace them with the notice
 * and other provisions required by the GPL or the LGPL. If you do not delete
 * the provisions above, a recipient may use your version of this file under
 * the terms of any one of the MPL, the GPL or the LGPL.
 *
 * ***** END LICENSE BLOCK ***** */

/**
 * Constants.
 */
const NS_ANNOTATIONS_SERVICE_CID = Components.ID("{4fd3212c-76f9-4322-9cbc-9f0a52c900e1}");
const NS_ANNOTATIONS_SERVICE_PROG_ID = "@mozilla.org/annotations_service;1";

// version
const NS_ANNOTATIONS_SERVICE_VERSION = "0.4.0.a.2";

function AnnotationsService() {
  this._servers = [];
  this._apps = {};
  this._autoLoadApps = [];
  this._autoLoadServers = [];
  this._autoloadSites = [];
  this._autoLoadUrl = null;
  this._nameCounter = 0;
  // pre-load the LocalServer
  var localServer = Components.classes["@mozilla.org/local_annotations_server;1"].getService();
}

/**
 * The URI for the LocalServer.
 */
AnnotationsService.LOCALSERVER_URI = "urn:localserver";
AnnotationsService.LOCALSERVER_USERNAME = "noUser";
AnnotationsService.LOCALSERVER_PASSWORD = "noPass";

AnnotationsService.prototype = {
  _servers: undefined,
  _autoLoadSites: undefined,
  _apps: undefined,
  _autoLoadApps: undefined,
  _autoLoadServers: undefined,
  _autoLoadUrl: undefined,
  _nameCounter: undefined
}; // end AnnotationsService.prototype
  


// PUBLIC XPCOM functions

/**
 * Get the version number for this Annotations Service.  The result
 * will be a string of the form (d.)+d  (for example, 1.2.1).
 * This function first appeared in version 0.2.2.
 *
 * @return version string
 */
AnnotationsService.prototype.getVersion = function() { return NS_ANNOTATIONS_SERVICE_VERSION; }

/**
 * Register an application.  The return value, an nsIAnnotationsApplication,
 * will be the portal for the application to access annotations.  
 * Applications are indexed by name--calling registerApplication twice
 * with the same name will return the same nsIAnnotationsApplication.
 *
 * @param name unique name of application
 * @exception if name is empty string
 */
AnnotationsService.prototype.registerApplication = function(name) {

    // if the name is null, autogenerate a name
    if ( (!name) || (name == "") ) {
	// generate a name while it is not unique
	do {
	    name = "_autogenerated_" + (this._nameCounter++);
	}
	while(this._apps[name]);
    }
    // else if the app name already exists, return it
    else if (this._apps[name]) {
	return this._apps[name].application;
    }
  
  this._apps[name] = {};

  var app = this._apps[name];

  app.servers = {};
  app.serverCount = 0;
  app.autoLoad = false;
  app.defaultPostServer = null;
  app.observers = [];
  app.observerAlerter = new ObserverAlerter(app);
  app.application = new AnnotationsApplication(name, this);

  return app.application.QueryInterface(Components.interfaces.nsIAnnotationsApplication);
}
  

/**
 * Unregister a previously registered application. Should only
 * be called once per registered application name.
 *
 * @param name name of previously registered application
 * @exception if name was not previously registered, or has already been unregistered
 */
AnnotationsService.prototype.unregisterApplication = function(name) {
  if (this._apps[name]) {
      var app = this._apps[name];

      // if autoLoad is on
      if (app.autoLoad) {
	// turn off autoLoad, cleaning up that stuff
	this._disableAutoLoadForApp(app);
      }

      for (uri in app.servers) {
	this._removeServerFromApp(app, uri);
      }
      
      delete this._apps[name];
  }
  else {
    throw Components.Exception("There is no application registered under the name " + name);
  }
}

AnnotationsService.prototype.QueryInterface = function(iid) {
  if (! (iid.equals(Components.interfaces.nsIAnnotationsService) ||
	 iid.equals(Components.interfaces.nsITrackerObserver) ||
	 iid.equals(Components.interfaces.nsISupports)) ) {
	   throw Components.results.NS_ERROR_NO_INTERFACE;
	 }
  return this;
}


// PRIVATE non-XPCOM functions for AnnotationsService

/**
 * Implementation of nsITrackerObserver.  Called when a page
 * begins loading.  Notifys the observers and begins the autoLoad, 
 * if necessary.
 * FIXED? -- if one observer exists, this may load annotations
 * for all servers, where that may be unnecessary
 */
AnnotationsService.prototype.onBeginLoad = function(url, win) {
  var doLoad = false;
  var allowSite = true;
  var sites = this._autoLoadSites;

  if (sites.length > 0) {
      allowSite = false;
      for (var ix=0; ix<sites.length; ix++) {
          if (url.indexOf(sites[ix]) == 0) {
              allowSite = true;
              dump("Autoload allowed for "+url+"\n");
              break;
          }
      }
  }

  // alert all autoLoad apps
  //  even on autoload rejection (eg. so they know to clear)
  for (var i = 0; i < this._autoLoadApps.length; i++) {
      // alert the app's observers
      var observers = this._autoLoadApps[i].observers;
      for (var j = 0; j < observers.length; j++) {
          observers[j].autoLoadOnBeginPageLoad(url, win);
	  // since we have at least one observer, we should do the load
          doLoad = true;
      }
  }

  if (!allowSite) {
      //dump("Autoload rejected for "+url+"\n");
      return; 
  } 
  
  // do the loads if necessary
  if (doLoad) {
    this._autoLoadUrl = url;
    for (var i = 0; i < this._autoLoadServers.length; i++) {
	var serverObject = this._autoLoadServers[i];
	var observers = new ObserverEnumerator(serverObject.autoLoadApps);
	serverObject.server.autoLoadAnnotations(url, observers, win);
    }
  }
}

AnnotationsService.prototype.onEndLoad = function(url, win) {
    // if we didn't do a load, no need to alert observers
    if (url != this._autoLoadUrl) {
	return;
    }
    
    for (var i = 0; i < this._autoLoadApps.length; i++) {
	// alert the app's observers
	var observers = this._autoLoadApps[i].observers;
	for (var j = 0; j < observers.length; j++) {
	    observers[j].autoLoadOnSuccessPageLoad(url, win);
	}
    }
}
  

AnnotationsService.prototype.onFailLoad = function(url, win) {
    // if we didn't do a load, no need to alert observers
    if (url != this._autoLoadUrl) {
	return;
    }
    
    for (var i = 0; i < this._autoLoadApps.length; i++) {
	// alert the app's observers
	var observers = this._autoLoadApps[i].observers;
	for (var j = 0; j < observers.length; j++) {
	    observers[j].autoLoadOnFailPageLoad(url, win);
	}
    }
}

AnnotationsService.prototype._addServer = function(appName, uri, username, password) {
  // test inputs
  var app = this._lookupAppName(appName);

  if ( (!uri) || (uri == "") ) {
    throw Components.Exception("URI blank or undefined.  Please use a valid server URI.");
  }
  
  // username and password for localServer are empty
  if (uri == AnnotationsService.LOCALSERVER_URI) {
    username = AnnotationsService.LOCALSERVER_USERNAME;
    password = AnnotationsService.LOCALSERVER_PASSWORD;
  }
  // non-localServer
  else {
    if (! username) {
      username = "";
    }
    if (! password) {
      password = "";
    }
  }

  // see if this server is already defined in the service
  var serverObject = this._findServer(uri, username, password);

  // if a server with this uri is already defined for our app
  if (app.servers[uri]) {
    /* if the service has no record of a server with this uri, username, and password,
       then we have changed the username and/or password of the server.  need to remove
       the server with old username & password from our app, and create the new username and password'ed server */
      if (! serverObject) {
	  // release the old server from our app
	  this._removeServerFromApp(app, uri);

	  // now, add the new server and link it up to our app
	  app.servers[uri] = this._addNewServer(uri, username, password, app);
      }
      /* otherwise, if there *is* a server by this uri, username, and password
	 extant in the annotationsservice, then either (1) we didn't change anything,
	 and are fine OR (2) we changed something, and another application already has
	 a server exactly like we need */
      // this is case (2) (case 1 requires no code)
      else if (serverObject.server != app.servers[uri].server) {
	  
	  // release the old server from our app
	  this._removeServerFromApp(app, uri);
	  
	  // add our app to the existing server
	  app.servers[uri] = this._addAppToExistingServer(serverObject, app);
      }
  } // end if the server exists in the app

  // otherwise, this server is new to the application
  // see if it is new to the service
  else {
    // if this server already exists in the service, add this app to it
    if (serverObject) {
      app.servers[uri] = this._addAppToExistingServer(serverObject, app);
    }
    // otherwise, create the new server and link it to our app
    else {
      app.servers[uri] = this._addNewServer(uri, username, password, app);
    }
  }

}

/**
 * Find a server.
 *
 * @param uri
 * @param username
 * @param password
 * @return serverObject, or null if not found
 */
AnnotationsService.prototype._findServer = function(uri, username, password) {
    /* run loop backwards, since recently added servers are more
       likely to be searched for (this function often called by add and remove) */
    for (var i = this._servers.length - 1; i >= 0; i--) {
	var server = this._servers[i].server;
	if ((server.uri == uri) && (server.username == username) && (server.password == password)) {
	    return this._servers[i];
	}
    }

    return null;
}

/**
 * Remove a server from an application.
 *
 * @param appName string name of app
 * @param uri string uri of server
 */
AnnotationsService.prototype._removeServer = function(appName, uri) {
    // test inputs
    var app = this._lookupAppName(appName);
    
    if ( (!uri) || (uri == "") ) {
	throw Components.Exception("URI blank or undefined.  Please use a valid URI.");
    }
    
    // find the server in the app list
    if (app.servers[uri]) {
	this._removeServerFromApp(app, uri);
    }
    else {
	throw Components.Exception("Server " + uri + " does not exist in " + appName + "'s list of servers.");
    }

    // if we've removed the default post server, make that change
    if (app.defaultPostServer && (app.defaultPostServer.url == uri)) {
      app.defaultPostServer = null;
    }

}


/**
 * Returns whether a server is defined for an application.
 *
 * @param appName string applications name
 * @param uri uri of server
 * @return whether a server of this uri is defined
 */
AnnotationsService.prototype._hasServer = function(appName, uri) {
    var app = this._lookupAppName(appName);

    return app.servers[uri] ? true : false;
}

/**
 *  Get the number of servers registered
 * for this app.
 *
 * @param appName string applications name
 * @return number of servers for appName
 */
AnnotationsService.prototype._getServerCount = function(appName) {
    var app = this._lookupAppName(appName);

    return app.serverCount;
}


/**
 * Set the default post server.
 *
 * @param appName string applications name
 * @param uri uri of server
 */
AnnotationsService.prototype._setDefaultPostServer = function(appName, uri) {
  var app = this._lookupAppName(appName);
    
  if (app.servers[uri]) {
    app.defaultPostServer = app.servers[uri].server;
  }
  else {
    throw new Components.Exception("Server " + uri + " does not exist for the application " + appName + ".");
  }
  
}
    
/**
 * Get the default post server.
 *
 * @param appName string application's name
 * @return string uri of default post server or null if no default post server defined
 */
AnnotationsService.prototype._getDefaultPostServer = function(appName) {
  var app = this._lookupAppName(appName);
  
  if (app.defaultPostServer) {
    return app.defaultPostServer.url;
  }
  else {
    return null;
  }

}
  
/**
 * Add an observer for an application.
 *
 * @param appName string name of application
 * @param observer nsIAnnotationsObserver
 */
AnnotationsService.prototype._addObserver = function(appName, observer) {
  var app = this._lookupAppName(appName);
  
  app.observers.push(observer);

}

/**
 * Remove an observer from an application.
 *
 * @param appName string name of application
 * @param observer nsIAnnotationsObserver
 */
AnnotationsService.prototype._removeObserver = function(appName, observer) {

  var app = this._lookupAppName(appName);
  var observers = app.observers;

  for (var i = 0; i < observers.length; i++) {
    if (observers[i] == observer) {
      observers.splice(i, 1);
      return;
    }
  }

  throw Components.Exception("The observer for which removal was requested does not exist for application " +
			     appName + ".");
}


/**
 * Get an enumerator with all server URIs for 
 * an application.
 *
 * @param appName string app name
 */
AnnotationsService.prototype._getServers = function(appName) {
  
    var app = this._lookupAppName(appName);
    
    var serversArray = [];
    for (uri in app.servers) {
      serversArray.push(uri);
    }
    
    return serversArray;
}
    
/**
 * Post an annotation.
 *
 * @param appName string
 * @param metadata nsIAnnotationMetadata
 * @param bodyText string
 * @param encoding string
 * @param postServerURL string
 */
AnnotationsService.prototype._postAnnotation = function(appName, metadata, bodyText, encoding, postServerURL) {
    
    var app = this._lookupAppName(appName);
    
    var server = null;
    if (postServerURL && (postServerURL != "")) {
	if (app.servers[postServerURL]) {
	    server = app.servers[postServerURL].server;
	}
	else {
	    app.observerAlerter.postAnnotationOnFailure("Requested post server: " + postServerURL + " does not exist for " +
							"application " + appName + ".");
	    return;
	}
    }
    else {
	if (app.defaultPostServer) {
	    server = app.defaultPostServer;
	}
	else {
	    app.observerAlerter.postAnnotationOnFailure("No default post server has been defined.");
	    return;
	}
    }

    server.postAnnotation(metadata, bodyText, encoding, app.observerAlerter);
}    

AnnotationsService.prototype._updateAnnotation = function(appName, annotation) {
    var app = this._lookupAppName(appName);

    if (! app.servers[annotation.serverName] ) {
	app.observerAlerter.updateAnnotationOnFailure(annotation.annotationURI,
						      "The server " + annotation.serverName + " does not exist " +
						      "in application " + appName + "'s list of servers.");
	return;
    }

    app.servers[annotation.serverName].server.updateAnnotation(annotation, app.observerAlerter);
}

AnnotationsService.prototype._deleteAnnotation = function(appName, annotation) {
   var app = this._lookupAppName(appName);

   if (! app.servers[annotation.serverName] ) {
     app.observerAlerter.deleteAnnotationOnFailure(annotation.annotationURI,
						   "The server " + annotation.serverName + " does not exist " +
						   "in application " + appName + "'s list of servers.");
     return;
   }

   app.servers[annotation.serverName].server.deleteAnnotation(annotation.annotationURI, app.observerAlerter);
}


AnnotationsService.prototype._getReplies = function(appName, annotation) {
   var app = this._lookupAppName(appName);

   if (! app.servers[annotation.serverName] ) {
     app.observerAlerter.getRepliesOnFailure(annotation,
					     "The server " + annotation.serverName + " does not exist " +
					     "in application " + appName + "'s list of servers.");
     return;
   }
   
   app.servers[annotation.serverName].server.getReplies(annotation, app.observerAlerter);
}

AnnotationsService.prototype._replyToAnnotation = function(appName, original, reply) {
  var app = this._lookupAppName(appName);
  
  if (! app.servers[original.serverName] ) {
    app.observerAlerter.replyToAnnotationOnFailure(original,
						   "The server " + original.serverName + " does not exist " +
						   "in application " + appName + "'s list of servers.");
    return;
  }
  
  app.servers[original.serverName].server.replyToAnnotation(original, reply, app.observerAlerter);
}

AnnotationsService.prototype._loadAnnotationBody = function(appName, annotation) {
  var app = this._lookupAppName(appName);
  
  if (! app.servers[annotation.serverName] ) {
      app.observerAlerter.loadAnnotationBodyOnFailure(annotation,
						      "The server " + annotation.serverName + " does not exist " +
						      "in application " + appName + "'s list of servers.");
      return;
  }
  
  app.servers[annotation.serverName].server.loadAnnotationBody(annotation, app.observerAlerter);
}

AnnotationsService.prototype._isAutoLoadEnabled = function(appName) {
  var app = this._lookupAppName(appName);

  return app.autoLoad;
}
    

AnnotationsService.prototype._autoLoadAnnotations = function(appName, loadAutomatically) {
  var app = this._lookupAppName(appName);
  
  // if turning autoLoad on
  if ((!app.autoLoad) && loadAutomatically) {
      app.autoLoad = true;

      // install all the backlinks from this app's servers back to the app
      for (aServer in app.servers) {
	var serverObj = app.servers[aServer];

	// add our app to autoLoadApps for this server
	serverObj.autoLoadApps.push(app);

	// if this server is new to autoLoading
	if (serverObj.autoLoadApps.length == 1) {
	  // add the server to the overall list of autoLoadServers
	    this._autoLoadServers.push(serverObj);
	}
      }

      // add this app to the list of autoLoadApps
      this._autoLoadApps.push(app);

      // if this is the first autoLoadApp, turn the tracker on
      if (this._autoLoadApps.length == 1) {
	  var tracker =  Components.classes["@mozilla.org/tracker;1"].getService();
	  tracker.QueryInterface(Components.interfaces.nsITrackerService);
	  tracker.addObserver(this);
      }
  }
  // if turning autoLoad off
  else if (app.autoLoad && (! loadAutomatically)) {
      this._disableAutoLoadForApp(app);
  }      
  
}
    
AnnotationsService.prototype._loadAnnotations = function(appName, url) {
    var app = this._lookupAppName(appName);
    
    for (serverName in app.servers) {
	app.servers[serverName].server.loadAnnotations(url, app.observerAlerter);
    }
}
     

/**
 * Disable autoLoad for an app.  If necessary,
 * turn off autoLoad for the AnnotationsService.
 *
 * @param app
 */
AnnotationsService.prototype._disableAutoLoadForApp = function(app) {
    app.autoLoad = false;
    
    for (aServer in app.servers) {
      var serverObj = app.servers[aServer];
      var autoLoadApps = serverObj.autoLoadApps;
      
      // remove the backlinks from the app's servers back to the app
      for (var i = 0; i < autoLoadApps.length; i++) {
	if (autoLoadApps[i] == app) {
	  autoLoadApps.splice(i, 1);
	  break;
	}
      }

      // if no more apps require autoLoading from this server
      if (autoLoadApps.length == 0) {
	// remove it from the autoLoad servers
	for (var i = 0; i < this._autoLoadServers.length; i++) {
	  if (this._autoLoadServers[i] == serverObj) {
	    this._autoLoadServers.splice(i, 1);
	    break;
	  }
	}
      }
    }

    // find the app in the _autoLoadApps array and remove it
    for (var i = 0; i < this._autoLoadApps.length; i++) {
      if (this._autoLoadApps[i] == app) {
	this._autoLoadApps.splice(i, 1);
	break;
      }
    }

    // if there are no more autoLoad apps, turn off the tracker
    if (this._autoLoadApps.length == 0) {
      var tracker = Components.classes["@mozilla.org/tracker;1"].getService();
      tracker.QueryInterface(Components.interfaces.nsITrackerService);
      tracker.removeObserver(this);
    }
}

      
/**
 * Adds an app to an existing server union object.
 *
 * @param serverObj server object
 * @param app app object
 * @return server object
 */
AnnotationsService.prototype._addAppToExistingServer = function(serverObj, app) {
  // bump the count for this server
  serverObj.allAppCount += 1;

  // if autoLoad is enabled for this app, create
  // a backlink to our app
  if (app.autoLoad) {
    serverObj.autoLoadApps.push(app);
    // if this is the first autoLoadApp for this server,
    // add it to list of autoLoadServers
    if (serverObj.autoLoadApps.length == 1) {
      this._autoLoadServers.push(serverObj);
    }
  }

  // increment count of total servers
  app.serverCount += 1;

  return serverObj;
}

/**
 * Adds a new server to the _servers data structure.
 * Assumes you have already tested that it does 
 * not exist with _findServer.
 *
 * @param uri 
 * @param username
 * @param password
 * @return new server union object 
 */
AnnotationsService.prototype._addNewServer = function(uri, username, password, app) {
  var newServer;

  // local server
  if (uri == AnnotationsService.LOCALSERVER_URI) {
    var newServer = Components.classes["@mozilla.org/local_annotations_server;1"].getService();
    newServer = newServer.QueryInterface(Components.interfaces.nsIAnnotationsServer);
  }
  // remote server
  else {
    var newServer = Components.classes["@mozilla.org/remote_annotations_server;1"].createInstance(Components.interfaces.nsIAnnotationsServer);
    newServer.init(uri, username, password);
  }

  var serverObj  = { server: newServer, allAppCount: 1, autoLoadApps: [] };
  
  if (app.autoLoad) {
    serverObj.autoLoadApps.push(app);
    this._autoLoadServers.push(serverObj);
  }
  
  // increment total # of servers in app
  app.serverCount += 1;

  // add the serverObject to the servers array
  this._servers.push(serverObj);
  
  return serverObj;
}


/**
 * Remove a server from an application.
 *
 * @param app Application 
 * @param uri string uri of server to release
 */
AnnotationsService.prototype._removeServerFromApp = function(app, uri) {
    
  // get the server object
  var serverObject = app.servers[uri];
  
  // if autoLoad is enabled for this app
  if (app.autoLoad) {
    // remove this app from the autoLoadApps for this server
    var autoLoadApps = serverObject.autoLoadApps;
    for (var i = 0; i < autoLoadApps.length; i++) {
      if (autoLoadApps[i] == app) {
	autoLoadApps.splice(i, 1);
	break;
      }
    }

    // if there are no more autoLoadApps in the server
    if (autoLoadApps.length == 0) {
      // remove this server from the _autoLoadServers
      for (var i = 0; i < this._autoLoadServers.length; i++) {
	  if (this._autoLoadServers[i] == serverObject) {
	      this._autoLoadServers.splice(i, 1);
	      break;
	  }
      }
    }
  } // end if app.autoLoad

  // decrease the allAppCount
  serverObject.allAppCount -= 1;
  
  // delete the server from the app's list
  delete app.servers[uri]; 

  // remove it from the total count of servers
  app.serverCount -= 1;

  // if there are no more apps using this server, delete it
  if (serverObject.allAppCount == 0) {
      for (var i = 0; i < this._servers.length; i++) {
	  if (this._servers[i] == serverObject) {
	      this._servers.splice(i, 1);
	      break;
	  }
      }
  }
}

AnnotationsService.prototype._lookupAppName = function(appName) {
  if (this._apps[appName]) {
    return this._apps[appName];
  }
  else {
    throw Components.Exception("Application " + appName + " has been unregistered and may no longer be used.");
  }
}

// AnnotationsService.prototype._enumerateDataStructures = function() {

//     dump("Enumerating _servers:\n");
//     for (var i = 0; i < this._servers.length; i++) {
// 	var serverObject = this._servers[i];
// 	var server = serverObject.server;
// 	dump(" - uri: " + server.url + "\n");
// 	dump(" - username: " + server.username + "\n");
// 	dump(" - password: " + server.password + "\n");
// 	dump(" - allAppCount: " + serverObject.allAppCount + "\n");
// 	dump(" - autoLoadApps array length: " + serverObject.autoLoadApps.length + "\n");
//     }

//     dump("Enumerating _apps:\n");
//     for (appName in this._apps) {
// 	var app = this._apps[appName];
// 	dump(" - appName: " + appName + "\n");
// 	dump("  - serverCount: " + app.serverCount + "\n");
// 	dump("  - autoLoad: " + app.autoLoad + "\n");
// 	for (uri in app.servers) {
// 	    dump ("  - server: " + uri + " with allAppCount: " + app.servers[uri].allAppCount + "\n");
// 	}
// 	if (app.defaultPostServer) {
// 	    dump("  - defaultPostServer: " + app.defaultPostServer.url + "\n");
// 	}
// 	dump("  - with " + app.observers.length + " observer(s)\n");
//     }
// }
 

/**
 * An nsISimpleEnumerator for observers.
 *
 * @param observers Array of nsIAnnotation
 */   
function ObserverEnumerator(apps) {
    this.apps = apps;
    this.nextElement = 0;
}

ObserverEnumerator.prototype.hasMoreElements = function() {
  return (this.nextElement < this.apps.length);
}

  /**
   * Returns the observerAlerter for each app.
   */
ObserverEnumerator.prototype.getNext = function() { 
    if (this.nextElement < this.apps.length) {
      var retVal = this.apps[this.nextElement].observerAlerter;
      this.nextElement += 1;
      return retVal;
    }
    else {
	throw Components.results.NS_ERROR_FAILURE;
    }
}

ObserverEnumerator.prototype.QueryInterface = function(iid) {
    if (!iid.equals(Components.interfaces.nsISimpleEnumerator) &&
	!iid.equals(Components.interfaces.nsISupports)) {
	    throw Components.results.NS_ERROR_NO_INTERFACE;
	}
    return this;
}

ObserverEnumerator.prototype.toString = function() { 
    return "[ObserverEnumerator]";
}

/**
 * Dispatches alerts events to all the 
 * observers of a particular application.
 * Handles all events except loads and autoLoads,
 * which are quite a bit more complicated.
 *
 * @param appObj object for the particular app
 */
function ObserverAlerter(appObj) {
    this.appObj = appObj;
}
    
ObserverAlerter.prototype = {
  QueryInterface: function(iid) {
    if (!iid.equals(Components.interfaces.nsIAnnotationsObserver) &&
	!iid.equals(Components.interfaces.nsISupports)) {
      throw Components.results.NS_ERROR_NO_INTERFACE;
    }
    return this;
  },
    postAnnotationOnSuccess: function(newAnnotation) {
      // assumes that the cost of cloning outweighs the cost
      // of an additional branch

      var observers = this.appObj.observers;
      var numObservers = observers.length;
      var lastObserver = numObservers - 1;
      
      for (var i = 0; i < numObservers; i++) {
	// observers 0 ... n - 1
	if (i != lastObserver) {
	  observers[i].postAnnotationOnSuccess(newAnnotation.clone());
	}
	// observer n, where we don't need to clone
	else {
	  observers[lastObserver].postAnnotationOnSuccess(newAnnotation);
	}
      }	
    },
    postAnnotationOnFailure: function(message) {
      var observers = this.appObj.observers;
      for (var i = 0; i < observers.length; i++) {
	  observers[i].postAnnotationOnFailure(message);
      }	
    },
    updateAnnotationOnSuccess: function(updatedAnnotation) {
      var observers = this.appObj.observers;
      var numObservers = observers.length;
      var lastObserver = numObservers - 1;
      
      for (var i = 0; i < numObservers; i++) {
	if (i != lastObserver) {
	  observers[i].updateAnnotationOnSuccess(updatedAnnotation.clone());
	}
	else {
	  observers[lastObserver].updateAnnotationOnSuccess(updatedAnnotation);
	}
      }	  	
    },
    updateAnnotationOnFailure: function(url, message) {
      var observers = this.appObj.observers;
      for (var i = 0; i < observers.length; i++) {
	observers[i].updateAnnotationOnFailure(url, message);
      }	
    },
    deleteAnnotationOnSuccess: function(url) {
      var observers = this.appObj.observers;
      for (var i = 0; i < observers.length; i++) {
	observers[i].deleteAnnotationOnSuccess(url);
      }	
    },
    deleteAnnotationOnFailure: function(url, message) {
      var observers = this.appObj.observers;
      for (var i = 0; i < observers.length; i++) {
	observers[i].deleteAnnotationOnFailure(url, message);
      }	
    },
  getRepliesOnSuccess: function(annotation, repliesLength, replies) { 
      var observers = this.appObj.observers;
      var numObservers = observers.length;
      var lastObserver = numObservers - 1;
      
      for (var i = 0; i < numObservers; i++) {
	  if (i != lastObserver) {
	    observers[i].getRepliesOnSuccess(annotation.clone(), replies.length,
					     this._cloneAnnotations(replies));
	  }
	  else {
	    observers[lastObserver].getRepliesOnSuccess(annotation, replies.length, 
							replies)
	      }
      }
  },
    getRepliesOnFailure: function(annotation, message) {
      var observers = this.appObj.observers;
      var numObservers = observers.length;
      var lastObserver = numObservers - 1;
      
      for (var i = 0; i < numObservers; i++) {
	if (i != lastObserver) {
	  observers[i].getRepliesOnFailure(annotation.clone(), message);
	}
	else {
	  observers[lastObserver].getRepliesOnFailure(annotation, message);
	}
      }	
    },
    replyToAnnotationOnSuccess: function(original, reply) {
      var observers = this.appObj.observers;
      var numObservers = observers.length;
      var lastObserver = numObservers - 1;
      
      for (var i = 0; i < numObservers; i++) {
	if (i != lastObserver) {
	  observers[i].replyToAnnotationOnSuccess(original.clone(), reply.clone());
	}
	else {
	  observers[lastObserver].replyToAnnotationOnSuccess(original, reply);
	}
      }	
    },
    replyToAnnotationOnFailure: function(original, message) {
      var observers = this.appObj.observers;
      var numObservers = observers.length;
      var lastObserver = numObservers - 1;
      
      for (var i = 0; i < numObservers; i++) {
	if (i != lastObserver) {
	  observers[i].replyToAnnotationOnFailure(original.clone(), message);
	}
	else {
	  observers[lastObserver].replyToAnnotationOnFailure(original, message);
	}
      }	
    },
    loadAnnotationBodyOnSuccess: function(annotation, bodyText) {
      var observers = this.appObj.observers;
      var numObservers = observers.length;
      var lastObserver = numObservers - 1;
      
      for (var i = 0; i < numObservers; i++) {
	if (i != lastObserver) {
	  observers[i].loadAnnotationBodyOnSuccess(annotation.clone(), bodyText);
	}
	else {
	  observers[lastObserver].loadAnnotationBodyOnSuccess(annotation, bodyText);
	}
      }	
    },
    loadAnnotationBodyOnFailure: function(annotation, message) {
      var observers = this.appObj.observers;
      var numObservers = observers.length;
      var lastObserver = numObservers - 1;
      
      for (var i = 0; i < numObservers; i++) {
	if (i != lastObserver) {
	  observers[i].loadAnnotationBodyOnFailure(annotation.clone(), message);
	}
	else {
	  observers[lastObserver].loadAnnotationBodyOnFailure(annotation, message);
	}
      }	
    },
    loadAnnotationsOnSuccess: function(serverURI, annotatedURI, annotationsLength, annotations) {
      var observers = this.appObj.observers;
      var numObservers = observers.length;
      var lastObserver = numObservers - 1;

      for (var i = 0; i < numObservers; i++) {
	if (i != lastObserver) {
	  observers[i].loadAnnotationsOnSuccess(serverURI, annotatedURI, 
						annotations.length,
						this._cloneAnnotations(annotations));
	}
	else {
	  observers[lastObserver].loadAnnotationsOnSuccess(serverURI, annotatedURI, 
							   annotations.length, annotations);
	}
      }
    },
    loadAnnotationsOnFailure: function(serverURI, annotatedURI, message) {
      var observers = this.appObj.observers;
      for (var i = 0; i < observers.length; i++) {
	  observers[i].loadAnnotationsOnFailure(serverURI, annotatedURI, message);
      }
    },
    autoLoadOnBeginPageLoad: function(url, win) {
      var observers = this.appObj.observers;
      for (var i = 0; i < observers.length; i++) {
	observers[i].autoLoadOnBeginPageLoad(url, win);
      }
    },
    autoLoadOnSuccessPageLoad: function(url, win) {
      var observers = this.appObj.observers;
      for (var i = 0; i < observers.length; i++) {
	observers[i].autoLoadOnSuccessPageLoad(url, win);
      }
    },
    autoLoadOnFailPageLoad: function(url, win) {
      var observers = this.appObj.observers;
      for (var i = 0; i <observers.length; i++) {
	  observers[i].autoLoadOnFailPageLoad(url, win);
      }
    },
  autoLoadAnnotationsForServer: function(serverURI, win, annotatedURI, arrayLength, annotations) {
      var observers = this.appObj.observers;
      var numObservers = observers.length;
      var lastObserver = numObservers - 1;
      
      for (var i = 0; i < numObservers; i++) {
	  if (i != lastObserver) {
	    observers[i].autoLoadAnnotationsForServer(serverURI, win, annotatedURI,
						      annotations.length, 
						      this._cloneAnnotations(annotations));
	  }
	  else {
	    observers[lastObserver].autoLoadAnnotationsForServer(serverURI, win, annotatedURI, 
								 annotations.length, 
								 annotations);
	  }
      }
  },
  autoLoadFailureForServer: function(serverURI, win, annotatedURI, message){
    var observers = this.appObj.observers;
    for (var i = 0; i < observers.length; i++) {
      observers[i].autoLoadFailureForServer(serverURI, win, annotatedURI, message);
    }
  },

  _extractAnnotations: function(annotEnum) {
    var annots = [];
    while(annotEnum.hasMoreElements()) {
      annots.push(annotEnum.getNext().QueryInterface(Components.interfaces.nsIAnnotation));
    }
    return annots;
  },
    
  _cloneAnnotations: function(oldAnnots) {
    var newAnnots = [];
    for (var i = 0; i < oldAnnots.length; i++) {
      newAnnots.push(oldAnnots[i].clone());
    }
    return newAnnots;
  }
};


function AnnotationsApplication(appname, service) {
  this._name = appname;
  this.service = service;
}

AnnotationsApplication.prototype = {
    get name() { return this._name; },
  get LOCALSERVER_URI() { return AnnotationsService.LOCALSERVER_URI; },

  // name of this application
  _name: undefined,
  // pointer to AnnotationsService
  service: undefined
};


AnnotationsApplication.prototype.QueryInterface = function(iid) {
  if (!iid.equals(Components.interfaces.nsIAnnotationsApplication) &&
      !iid.equals(Components.interfaces.nsISupports)) {
	throw Components.results.NS_ERROR_NO_INTERFACE;
      }
  return this;
}


/**
 * Add an observer that implements callback
 * functions to receive the results of annotations 
 * calls.
 *
 * @param observer nsIAnnotationsObserver
 */
AnnotationsApplication.prototype.addObserver = function(observer) {
  this.service._addObserver(this._name, observer);
}
  

/**
 * Remove a previously added observer.
 *
 * @param observer nsIAnnotationsObserver
 * @exception if observer was not previously added
 */
AnnotationsApplication.prototype.removeObserver = function(observer) {
  this.service._removeObserver(this._name, observer);
}

  
/**
 * Add an annotations server to this application's
 * collection of servers. If a server of this url 
 * with different username and password already exists,
 * the new username and password will be used. 
 *
 * @param uri string uri of server
 * @param username string
 * @param password string 
 */
AnnotationsApplication.prototype.addServer = function(uri, username, password) {
  this.service._addServer(this._name, uri, username, password);
}

/**
 * Remove a previously added server from this
 * application's collection of servers.
 *
 * @param url string url of previously added server
 * @exception if server is not in this application's collection of servers
 */
AnnotationsApplication.prototype.removeServer = function(url) {
  this.service._removeServer(this._name, url);
}

/**
 * Returns whether a server is defined for this application.
 *
 * @param uri string uri to test whether exists
 * @return whether a server of this uri is defined
 */
AnnotationsApplication.prototype.hasServer = function(uri) {
    return this.service._hasServer(this._name, uri);
}

/**
 * Get the number of servers registered
 * for this app.
 *
 * @return number of servers for this app
 */
AnnotationsApplication.prototype.getServerCount = function() {
    return this.service._getServerCount(this._name);
}

/**
 * Set the server that the postAnnotation operation
 * will use by default.
 *
 * @param uri string default server uri
 */
AnnotationsApplication.prototype.setDefaultPostServer = function(uri) {
  this.service._setDefaultPostServer(this._name, uri);
}

  /**
   * Get the default server used by postAnnotation.
   * 
   * @return url of default server or null if no default server has been set
   */
AnnotationsApplication.prototype.getDefaultPostServer = function() {
  return this.service._getDefaultPostServer(this._name);
}

/**
 * Get a list of the URIs of all 
 * servers for this application.
 *
 * @param serversLength pass in a blank Object
 * @param servers Object whose "value" field will contain
 *                an array of server URIs upon return
 */ 
AnnotationsApplication.prototype.getServers = function(serversLengthObj, serversObj) {
  serversObj.value = this.service._getServers(this._name);
  serversLengthObj.value = serversObj.value.length;
}

/**
 *
 * Returns whether autoLoad is enabled.
 *
 * @return boolean whether autoLoad is enabled
 */
AnnotationsApplication.prototype.isAutoLoadEnabled = function() {
    return this.service._isAutoLoadEnabled(this._name);
}

/**
 * Set whether this application wants automatic
 * loading of annotations.  Automatic loading of
 * annotations means that whenever the user loads
 * a new page in the browser, the application will
 * be updated with that page's URL.  Additionally,
 * all servers for this application will be queried
 * for annotations applying to the new URL.  These
 * annotations will be offered to registered observers
 * via the callback functions prefixed with autoLoad
 * in the nsIAnnotationsObserver interface.
 *
 * If autoLoad is not set by this function,
 * it defaults to false -- ie, do NOT load automatically.
 *
 * @param loadAutomatically boolean whether to load annotations automatically
 */
AnnotationsApplication.prototype.autoLoadAnnotations = function(loadAutomatically) {
    this.service._autoLoadAnnotations(this._name, loadAutomatically);
}

/*
 * Set sites to allow for autoload, empty to allow all
 *
 * @param siteString space delimited list of sites to filter
 */
AnnotationsApplication.prototype.autoLoadSites = function(siteString) {
    //should have access function for this
    this.service._autoLoadSites = siteString.split(" ");
}

/**
 * Post an annotation to an annotations server asynchronously.  If no
 * postServerName is provided, then the annotation will be posted to the
 * default annotations server defined by setDefaultPostServer.  
 * This function returns immediately.  Results are provided to registered
 * observers via the callback functions:
 * 
 *  void postAnnotationOnSuccess(in nsIAnnotation newAnnotation);
 *  void postAnnotationOnFailure(in string message);
 * which are part of the nsIAnnotationsObserver interface.
 *
 * @param metadata nsIAnnotationMetadata object for the annotations
 * @param bodyText string text of the body of the annotation
 * @param encoding character encoding for the annotation, or null for no
 *                         specified encoding
 * @param postServerURL string optional url of postServer to use. this
 *                      postServer must have been added by calling addServer.
 *                      if postServerURL is null or "", the default post server 
 *                      will be used
 * @exception if postServerName is invalid, or if relying on default when no
 *            default has been defined
 */
AnnotationsApplication.prototype.postAnnotation = function(metadata, bodyText, encoding, postServerURL) {
    this.service._postAnnotation(this._name, metadata, bodyText, encoding, postServerURL);
}
    

/**
 * Updates an annotation asynchronously on the server where it originated. 
 * Requires that this server be part of the collection of servers for
 * this application (that is, it must have been added by addServer).
 * This function returns immediately.  Results are provided to registered
 * observers via the callback functions:
 * 
 *   void updateAnnotationOnSuccess(in nsIAnnotation updatedAnnotation);
 *   void updateAnnotationOnFailure(in string url, in string message);
 * which are part of the nsIAnnotationsObserver interface.
 *
 * @param annotation nsIAnnotation to be updated on server
 */
AnnotationsApplication.prototype.updateAnnotation = function(annotation) {
    this.service._updateAnnotation(this._name, annotation);
}

/**
 * Deletes an annotation asynchronously from the server where it originated.
 * That server must be defined--that is, you must have called addServer for it.
 * This function returns immediately.  Results are provided to registered
 * observers via the callback functions:
 * 
 *   void deleteAnnotationOnSuccess(in string url);
 *   void deleteAnnotationOnFailure(in string url, in string message);
 * which are part of the nsIAnnotationsObserver interface.
 *
 * @param annotation nsIAnnotation to delete from server
 */
AnnotationsApplication.prototype.deleteAnnotation = function(annotation) {
    this.service._deleteAnnotation(this._name, annotation);
}

/**
 * Load the replies for an annotation asynchronously.
 * This function returns immediately.  Results are provided to registered
 * observers via the callback functions:
 * 
 *   void getRepliesOnSuccess(in nsIAnnotation annotation, in nsISimpleEnumerator repliesEnum);
 *   void getRepliesOnFailure(in nsIAnnotation annotation, in string message);
 * which are part of the nsIAnnotationsObserver interface.
 *
 * @param annotation nsIAnnotation for which to load replies
 */
AnnotationsApplication.prototype.getReplies = function(annotation) {
    this.service._getReplies(this._name, annotation);
}
    
/**
 * Reply to an annotation on an annotations server asynchronously.  
 * This function returns immediately.  Results are provided to registered
 * observers via the callback functions:
 * 
 *  void replyToAnnotationOnSuccess(in nsIAnnotation original, in nsIAnnotation reply);
 *  void replyToAnnotationOnFailure(in nsIAnnotation original, in string message);
 * which are part of the nsIAnnotationsObserver interface.
 *
 * @param original nsIAnnotation to reply to
 * @param reply nsIAnnotationMetadata for the annotation to create
 */
AnnotationsApplication.prototype.replyToAnnotation = function(original, reply) {
    this.service._replyToAnnotation(this._name, original, reply);
}


/**
 * Load the annotations for a page asynchronously.  This function may be used 
 * for one-time loads, or to load annotations for a page that
 * the user is not currently viewing.  To load annotations
 * for each page that the user views, see the autoLoadAnnotations
 * function above.
 *
 * This function returns immediately.  Results are provided to registered
 * observers via the callback functions:
 *
 *  void loadAnnotationsOnTimeAnnotations(in string url, in nsISimpleEnumerator onTimeAnnotations);
 *  void loadAnnotationsLateAnnotations(in string url, in nsISimpleEnumerator lateAnnotations);
 *  void loadAnnotationsOnFailure(in string url, in string message);
 *
 * @param url string URL to load annotations for
 */
AnnotationsApplication.prototype.loadAnnotations = function(url) {
    this.service._loadAnnotations(this._name, url);
}
  

/**
 * Load the body text of an annotation asynchronously from the server.  
 *
 * This function returns immediately.  Results are provided to registered
 * observers via the callback functions:
 * 
 *   void loadAnnotationBodyOnSuccess(in nsIAnnotation annotation, in string bodyText);
 *   void loadAnnotationBodyOnFailure(in nsIAnnotation annotation, in string message);
 *
 * @param annotation nsIAnnotation to have body loaded for
 */
AnnotationsApplication.prototype.loadAnnotationBody = function(annotation) {
    this.service._loadAnnotationBody(this._name, annotation);
}











/**************************
 * XPCOM stuff
 **************************/

/**
 * Factory for nsAnnotationsService.
 */
var annotServiceFactory = new Object();

annotServiceFactory.createInstance = function (outer, iid) {
    if (outer != null) {
	throw Components.results.NS_ERROR_NO_AGGREGATION;
    }
    
    if (!iid.equals(Components.interfaces.nsISupports)) {
	throw Components.results.NS_ERROR_INVALID_ARG;
    }

    return new AnnotationsService().QueryInterface(iid);
}


/**
 * Module for nsAnnotationsService.
 */
var nsAnnotationsServiceModule = new Object();

nsAnnotationsServiceModule.registerSelf =
    function (compMgr, fileSpec, location, type) {
	compMgr = compMgr.QueryInterface(Components.interfaces.nsIComponentRegistrar);
	dump("***** Annotations Service -- registering self ******\n");
	debug("***** Annotations Service -- registering self ******\n");
	compMgr.registerFactoryLocation(NS_ANNOTATIONS_SERVICE_CID,
					"Annotations Service",
					NS_ANNOTATIONS_SERVICE_PROG_ID, 
					fileSpec,
					location,
					type);
    }

nsAnnotationsServiceModule.unregisterSelf = function(compMgr, fileSpec, location) {
    compMgr = compMgr.QueryInterface(Components.interfaces.nsIComponentRegistrar);
    compMgr.unregisterFactoryLocation(NS_ANNOTATIONS_SERVICE_CID, fileSpec);
}

nsAnnotationsServiceModule.getClassObject = function (compMgr, cid, iid) {
    if (cid.equals(NS_ANNOTATIONS_SERVICE_CID)) {
	return annotServiceFactory;
    }
    else {
	throw Components.results.NS_ERROR_NO_INTERFACE;  
    }
}

nsAnnotationsServiceModule.canUnload = function(compMgr) {
    return true;
}

/* entrypoint */
function NSGetModule(compMgr, fileSpec) {
    return nsAnnotationsServiceModule;
}



